/*
 * vi:ts=4:tw=78:shiftwidth=4:expandtab
 * vim600:fdm=marker
 *
 * bpsf.c  -  useful routines for manipulating BPSF font file
 *
 * Copyright (C) 2004 by ejoy <ejoy@users.sourceforge.net>
 * Begin       : 27-Jul-2004
 * Last Change : 10-May-2006.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser GPL (LGPL) as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "bpsf.h"

bpsf_info* bpsf_create() {
    bpsf_info* info;
    int i;
    info = malloc(sizeof(bpsf_info));
    assert(info);
    for (i = 0; i < 65536; ++i)
        info->m_unicode_map[i] = (unsigned short)(-1);
    info->m_filename   = 0;
    info->m_bitmap     = 0;
    info->m_width      = 0;
    info->m_height     = 0;
    info->m_glyph_size = 0;
    info->m_glyph_count = 0;
    info->m_use_mmap   = 0;
    return info;
}

/* TODO: mmap support */
int bpsf_open(bpsf_info* info, const char* file) {
/*    info->m_use_mmap = use_mmap;*/
    int   rc;
    int   bitmap_size;
    FILE *fp;
    char  buf[9];

    fp = fopen(file, "rb");
    if (fp == 0) {
        perror("failed to open font file");
        return -1;
    }

    if (fread(buf, 9, 1, fp) != 1) { /* header is 9 bytes,  no padding */
        perror("failed to read font header");
        fclose(fp);
        return -1;
    }

    if (!(buf[0] == 0x36 && buf[1] == 0x04)) {
        perror("this is not a bpsf file (wrong magic number)");
        fclose(fp);
        return -1;
    }

    if (buf[2] != 5) {
        perror("only bpsf file with unicode data (filemode = 5) is support");
        fclose(fp);
        return -1;
    }

    /* file header ok, now read on */
    info->m_width       = buf[3];
    info->m_height      = buf[4];
    info->m_glyph_size  = info->m_width * info->m_height / 8;
    info->m_glyph_count = ((unsigned*)(buf + 5))[0];

    assert(info->m_filename == 0);
    info->m_filename = malloc(strlen(file) + 1);
    assert(info->m_filename);
    strcpy(info->m_filename, file);

    /* read in bitmap data */
    bitmap_size = info->m_glyph_size * info->m_glyph_count;
    info->m_bitmap = malloc(bitmap_size);
    assert(info->m_bitmap);
    if (fread(info->m_bitmap, bitmap_size, 1, fp) != 1) {
        perror("failed to load bitmap data");
        free(info->m_filename);
        free(info->m_bitmap);
        fclose(fp);
        return -1;
    }

    /* read in unicode map */
    {
        unsigned i;
        unsigned short buf[2];
        for (i = 0; i < info->m_glyph_count; ++i) {
            rc = fread(&buf, sizeof(buf), 1, fp);
            assert(rc == 1);
            assert(buf[1] = 0xffff); /* we assume each glyph's unicode value is a ucs16 integer */
            info->m_unicode_map[buf[0]] = i;
        }
    }

    fclose(fp);
    return 0;
}

/* TODO: mmap support */
int bpsf_destory(bpsf_info* info) {
    assert(info->m_filename);
    assert(info->m_bitmap);
    free(info->m_filename);
    free(info->m_bitmap);
    free(info);
    return 0;
}

unsigned char* bpsf_index2(bpsf_info* info, unsigned short c) { /* ucs16 */
        return &info->m_bitmap[c * info->m_glyph_size];
}

unsigned char* bpsf_index(bpsf_info* info, unsigned short c) { /* ucs16 */
    unsigned short index;
    index = info->m_unicode_map[c];
    if (index != (unsigned short)(-1))
        return &info->m_bitmap[index * info->m_glyph_size];
    else
        return 0;
}

/* TODO: unified utype */
typedef unsigned short Uint16;

static Uint16 utf8_to_unicode(const unsigned char *utf8) {
    Uint16 ch = utf8[0];

    if ( ch >= 0xF0 ) {
        ch  =  (Uint16)((*utf8++)&0x07) << 18;
        ch |=  (Uint16)((*utf8++)&0x3F) << 12;
        ch |=  (Uint16)((*utf8++)&0x3F) << 6;
        ch |=  (Uint16)((*utf8)&0x3F);
    } else if ( ch >= 0xE0 ) {
        ch  =  (Uint16)((*utf8++)&0x3F) << 12;
        ch |=  (Uint16)((*utf8++)&0x3F) << 6;
        ch |=  (Uint16)((*utf8)&0x3F);
    } else if ( ch >= 0xC0 ) {
        ch  =  (Uint16)((*utf8++)&0x3F) << 6;
        ch |=  (Uint16)((*utf8)&0x3F);
    }
    return ch;
}

unsigned char* bpsf_index_utf8(bpsf_info* info, const unsigned char
        *utf8) {
    return bpsf_index(info, utf8_to_unicode(utf8));
}


